;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "auto-complete" "20251231.1622"
  "Auto Completion for GNU Emacs."
  '((emacs "25.1")
    (popup "0.5.8"))
  :url "https://github.com/auto-complete/auto-complete"
  :commit "07f9915e08342410b933145d7934998709753a29"
  :revdesc "07f9915e0834"
  :keywords '("completion" "convenience")
  :authors '(("Tomohiro Matsuyama" . "m2ym.pub@gmail.com"))
  :maintainers '(("Jen-Chieh Shen" . "jcs090218@gmail.com")))
